(*
 * <COPYRIGHT>
 * Copyright 2005 Gerd Stolpmann
 *
 * <GPL>
 * This file is part of UMLMON.
 *
 * UMLMON is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * UMLMON is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with UMLMON; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * </>
 *)

(* $Id: termemu.ml 187 2005-10-14 10:54:15Z gerd $ *)

(* Terminal emulation for Linux consoles
 *
 * See console_codes(4) for a list of control characters. ECMA-48 (freely available)
 * may also be a source of information.
 *)

open Printf


type intensity = I_norm | I_dim | I_bold

type color = Black | Red | Green | Brown | Blue | Magenta | Cyan | White

let color =
  function
    | Black -> "black"
    | Red -> "red"
    | Green -> "green"
    | Brown -> "brown"
    | Blue -> "blue"
    | Magenta -> "magenta"
    | Cyan -> "cyan"
    | White -> "white"

type atts =
    { intense : intensity;
      underscore : bool;
      blink : bool;
      reverse : bool;
      fg_color : color;
      bg_color : color
    }
  (* Attribute components are intentionally immutable, so atts can only be updated
   * as a whole.
   *)

type mapping = ISO | Graphics | CP437  (* Other not supported *)

type charset = Mapping of mapping | UTF8 of mapping  (* saved former mapping *)

type cell_char = Empty | Unicode of int

type key =
    [ `Translated of int
	(* An already translated Unicode key *)
    | `Cursor_up
    | `Cursor_down
    | `Cursor_left
    | `Cursor_right
    | `Page_up
    | `Page_down
    | `Home
    | `End
    | `Insert
    | `Delete
    | `Backspace
    | `Enter
    | `Tab
    | `Escape
    | `Fn of int
	(* Function key. int = 1..12 *)
    ]


exception Found of bool * (unit -> unit)


let csi_re =
  Str.regexp "\027\\[\\??\\([0-9]*\\)\\(;[0-9]*\\)?\\(;[0-9]*\\)?\\(;[0-9]*\\)?\\(;[0-9]*\\)?\\(;[0-9]*\\)?\\(;[0-9]*\\)?\\(;[0-9]*\\)?\\(;[0-9]*\\)?\\(;[0-9]*\\)?\\(;[0-9]*\\)?\\(;[0-9]*\\)?\\(;[0-9]*\\)?\\(;[0-9]*\\)?\\(;[0-9]*\\)?\\(;[0-9]*\\)?[\064-\090\092-\126]$"

let get_csi_params s =
  let rec loop k =
    try
      let g = Str.matched_group k s in
      let n =
	if g = "" || g = ";" then
	  0
	else
	  try
	    if g.[0] = ';' then
	      int_of_string(String.sub g 1 (String.length g - 1))
	    else
	      int_of_string g 
	  with
	      Failure _ -> max_int in
      n :: loop (k+1)
    with
	Not_found -> []
  in
  loop 1
;;


let default_atts =
  { intense = I_norm;
    underscore = false;
    blink = false;
    reverse = false;
    fg_color = White;
    bg_color = Black 
  }

let rc = 0x00D7 ;;   (* replacement character = multiplication sign *)

let graphics_mapping =
  [| 0x0000; 0x0001; 0x0002; 0x0003; 0x0004; 0x0005; 0x0006; 0x0007;
     0x0008; 0x0009; 0x000a; 0x000b; 0x000c; 0x000d; 0x000e; 0x000f;
     0x0010; 0x0011; 0x0012; 0x0013; 0x0014; 0x0015; 0x0016; 0x0017;
     0x0018; 0x0019; 0x001a; 0x001b; 0x001c; 0x001d; 0x001e; 0x001f;
     0x0020; 0x0021; 0x0022; 0x0023; 0x0024; 0x0025; 0x0026; 0x0027;
     0x0028; 0x0029; 0x002a; 0x2192 (*ok*); 0x2190 (*ok*); 0x2191 (*ok*); 0x2193 (*ok*); 0x002f;
     (* 2190-2193: thin arrows, defined for HTML-4 *)
     (-1) (* 0x2588 *); 0x0031; 0x0032; 0x0033; 0x0034; 0x0035; 0x0036; 0x0037;
     (* 2588: full block. Emulated by reverse video *)
     0x0038; 0x0039; 0x003a; 0x003b; 0x003c; 0x003d; 0x003e; 0x003f;
     0x0040; 0x0041; 0x0042; 0x0043; 0x0044; 0x0045; 0x0046; 0x0047;
     0x0048; 0x0049; 0x004a; 0x004b; 0x004c; 0x004d; 0x004e; 0x004f;
     0x0050; 0x0051; 0x0052; 0x0053; 0x0054; 0x0055; 0x0056; 0x0057;
     0x0058; 0x0059; 0x005a; 0x005b; 0x005c; 0x005d; 0x005e; 0x00a0;
     0x2666 (* 0x25c6 *); rc (* 0x2592 *); rc (* 0x2409 *); rc (* 0x240c *); rc (* 0x240d *); rc (* 0x240a *); 0x00b0; 0x00b1;
     rc (* 0x2591 *); rc (* 0x240b *); rc (* 0x2518 *); rc (* 0x2510 *); rc (* 0x250c *); rc (* 0x2514 *); rc (* 0x253c *); rc (* 0xf800 *);
     rc (* 0xf801*); 0x2014 (* 0x2500 *); rc (* 0xf803 *); rc (* 0xf804 *); rc (* 0x251c *); rc (* 0x2524 *); rc (* 0x2534 *); rc (* 0x252c *);
     (* 2500: horizontal line, replaced by mdash *)
     0x007C (*0x2502*); 0x2264 (* ok *); 0x2265 (* ok *); 0x03c0 (* ok *); 0x2260 (* ok *); 0x00a3; 0x00b7; 0x007f;
     (* 0x2502: vertical bar *)
     0x0080; 0x0081; 0x0082; 0x0083; 0x0084; 0x0085; 0x0086; 0x0087;
     0x0088; 0x0089; 0x008a; 0x008b; 0x008c; 0x008d; 0x008e; 0x008f;
     0x0090; 0x0091; 0x0092; 0x0093; 0x0094; 0x0095; 0x0096; 0x0097;
     0x0098; 0x0099; 0x009a; 0x009b; 0x009c; 0x009d; 0x009e; 0x009f;
     0x00a0; 0x00a1; 0x00a2; 0x00a3; 0x00a4; 0x00a5; 0x00a6; 0x00a7;
     0x00a8; 0x00a9; 0x00aa; 0x00ab; 0x00ac; 0x00ad; 0x00ae; 0x00af;
     0x00b0; 0x00b1; 0x00b2; 0x00b3; 0x00b4; 0x00b5; 0x00b6; 0x00b7;
     0x00b8; 0x00b9; 0x00ba; 0x00bb; 0x00bc; 0x00bd; 0x00be; 0x00bf;
     0x00c0; 0x00c1; 0x00c2; 0x00c3; 0x00c4; 0x00c5; 0x00c6; 0x00c7;
     0x00c8; 0x00c9; 0x00ca; 0x00cb; 0x00cc; 0x00cd; 0x00ce; 0x00cf;
     0x00d0; 0x00d1; 0x00d2; 0x00d3; 0x00d4; 0x00d5; 0x00d6; 0x00d7;
     0x00d8; 0x00d9; 0x00da; 0x00db; 0x00dc; 0x00dd; 0x00de; 0x00df;
     0x00e0; 0x00e1; 0x00e2; 0x00e3; 0x00e4; 0x00e5; 0x00e6; 0x00e7;
     0x00e8; 0x00e9; 0x00ea; 0x00eb; 0x00ec; 0x00ed; 0x00ee; 0x00ef;
     0x00f0; 0x00f1; 0x00f2; 0x00f3; 0x00f4; 0x00f5; 0x00f6; 0x00f7;
     0x00f8; 0x00f9; 0x00fa; 0x00fb; 0x00fc; 0x00fd; 0x00fe; 0x00ff |]

let cp437_mapping =
  [| 0x0000; rc (* 0x263a *); rc (* 0x263b *); 0x2665 (* ok *); 0x2666 (* ok *); 0x2663 (* ok *) ; 0x2660 (* ok *); 0x2022 (* ok *);
     rc (* 0x25d8 *); rc (* 0x25cb *); rc (* 0x25d9 *); rc (* 0x2642 *); rc (* 0x2640 *); rc (* 0x266a *); rc (* 0x266b *); rc (* 0x263c *);
     rc (* 0x25b6 *); rc (* 0x25c0 *); rc (* 0x2195 *); rc (* 0x203c *); 0x00b6; 0x00a7; rc (* 0x25ac *); rc (* 0x21a8 *);
     0x2191 (* ok *); 0x2193 (* ok *); 0x2192 (* ok *); 0x2190 (* ok *); 0x2308 (* 0x221f *); rc (* 0x2194 *); rc (* 0x25b2 *); rc (* 0x25bc *);
     (* 221F: right angle, replaced by lceil *)
     0x0020; 0x0021; 0x0022; 0x0023; 0x0024; 0x0025; 0x0026; 0x0027;
     0x0028; 0x0029; 0x002a; 0x002b; 0x002c; 0x002d; 0x002e; 0x002f;
     0x0030; 0x0031; 0x0032; 0x0033; 0x0034; 0x0035; 0x0036; 0x0037;
     0x0038; 0x0039; 0x003a; 0x003b; 0x003c; 0x003d; 0x003e; 0x003f;
     0x0040; 0x0041; 0x0042; 0x0043; 0x0044; 0x0045; 0x0046; 0x0047;
     0x0048; 0x0049; 0x004a; 0x004b; 0x004c; 0x004d; 0x004e; 0x004f;
     0x0050; 0x0051; 0x0052; 0x0053; 0x0054; 0x0055; 0x0056; 0x0057;
     0x0058; 0x0059; 0x005a; 0x005b; 0x005c; 0x005d; 0x005e; 0x005f;
     0x0060; 0x0061; 0x0062; 0x0063; 0x0064; 0x0065; 0x0066; 0x0067;
     0x0068; 0x0069; 0x006a; 0x006b; 0x006c; 0x006d; 0x006e; 0x006f;
     0x0070; 0x0071; 0x0072; 0x0073; 0x0074; 0x0075; 0x0076; 0x0077;
     0x0078; 0x0079; 0x007a; 0x007b; 0x007c; 0x007d; 0x007e; rc (* 0x2302 *);
     0x00c7; 0x00fc; 0x00e9; 0x00e2; 0x00e4; 0x00e0; 0x00e5; 0x00e7;
     0x00ea; 0x00eb; 0x00e8; 0x00ef; 0x00ee; 0x00ec; 0x00c4; 0x00c5;
     0x00c9; 0x00e6; 0x00c6; 0x00f4; 0x00f6; 0x00f2; 0x00fb; 0x00f9;
     0x00ff; 0x00d6; 0x00dc; 0x00a2; 0x00a3; 0x00a5; rc (* 0x20a7 *); 0x0192 (* ok *);
     0x00e1; 0x00ed; 0x00f3; 0x00fa; 0x00f1; 0x00d1; 0x00aa; 0x00ba;
     0x00bf; rc (* 0x2310 *); 0x00ac; 0x00bd; 0x00bc; 0x00a1; 0x00ab; 0x00bb;
     rc (* 0x2591 *); rc (* 0x2592 *); rc (* 0x2593 *); 0x007c (* 0x2502 *); rc (* 0x2524 *); rc (* 0x2561 *); rc (* 0x2562 *); rc (* 0x2556 *);
     (* 2502: vertical bar *)
     rc (* 0x2555 *); rc (* 0x2563 *); rc (* 0x2551 *); rc (* 0x2557 *); rc (* 0x255d *); rc (* 0x255c *); rc (* 0x255b *); rc (* 0x2510 *);
     rc (* 0x2514 *); rc (* 0x2534 *); rc (* 0x252c *); rc (* 0x251c *); 0x2D (* 0x2500 *); rc (* 0x253c *); rc (* 0x255e *); rc (* 0x255f *);
     (* 2500: horizontal bar *)
     rc (* 0x255a *); rc (* 0x2554 *); rc (* 0x2569 *); rc (* 0x2566 *); rc (* 0x2560 *); 0x3D (* 0x2550 *); rc (* 0x256c *); rc (* 0x2567 *); 
     (* 2550: similar to equal sign *)
     rc (* 0x2568 *); rc (* 0x2564 *); rc (* 0x2565 *); rc (* 0x2559 *); rc (* 0x2558 *); rc (* 0x2552 *); rc (* 0x2553 *); rc (* 0x256b *);
     rc (* 0x256a *); rc (* 0x2518 *); rc (* 0x250c *); (-1) (* 0x2588 *); rc (* 0x2584 *); rc (* 0x258c *); rc (* 0x2590 *); rc (* 0x2580 *);
     0x03b1 (* ok *); 0x00df; 0x0393 (* ok *); 0x03c0 (* ok *) ; 0x03a3 (* ok *); 0x03c3 (* ok *); 0x00b5; 0x03c4 (* ok *);
     0x03a6 (* ok *); 0x0398 (* ok *); 0x03a9 (* ok *); 0x03b4 (* ok *); 0x221e (* ok *); 0x03c6 (* ok *); 0x03b5 (* ok *); 0x2229 (* ok *);
     0x2261 (* ok *); 0x00b1; 0x2265 (* ok *); 0x2264 (* ok *); rc (* 0x2320 *); rc (* 0x2321 *); 0x00f7; 0x2248 (* ok *);
     0x00b0; 0x22c9 (* 0x2219 *); 0x00b7; 0x221a (* ok *); rc (* 0x207f *); 0x00b2; rc (* 0x25a0 *); 0x00a0 |]
     (* 2219: small bullet *)




let map_char mapping c =
  match mapping with
    | ISO      -> Char.code c
    | Graphics -> graphics_mapping.( Char.code c )
    | CP437    -> cp437_mapping.( Char.code c )


class terminal_emulator width height =
object(self)

  (* --- Screen state, modes, attributes --- *)

  val mutable screen = Array.init height 
			 (fun _ -> Array.init width
			    (fun _ -> (Empty, default_atts)))
  val mutable scroll_first = 0
  val mutable scroll_last = height-1   (* Scrollable region *)
  val mutable tab_stops = Array.make width false  (* Tab stops, initialized later *)
  val mutable x = 0
  val mutable y = 0   (* cursor location *)
  val mutable atts =  default_atts
  val mutable disp_ctrl = false     (* Whether to display control characters *)
  val mutable toggle_meta = false   (* Toggle bit 7 *)
  val mutable insert_mode = false
  val mutable decckm = false        (* About cursor keys *)
  val mutable rel_addr = false      (* Addressing relative to scrolling region *)
  val mutable autowrap = true
  val mutable cur_visible = true
  val mutable fg_default = White
  val mutable bg_default = Black
  val mutable charset = Mapping ISO
  val mutable g0_mapping = ISO
  val mutable g1_mapping = Graphics

  val mutable need_wrap = false   (* Whether to output CR/LF before the next char *)

  val mutable saved_x = 0
  val mutable saved_y = 0   (* cursor location *)
  val mutable saved_atts = default_atts
  val mutable saved_charset = Mapping ISO
  val mutable saved_g0_mapping = ISO
  val mutable saved_g1_mapping = ISO

  (* --- output side buffers --- *)

  val mutable esc_seq = Buffer.create 50
     (* The current ESC sequence. CSI is translated into ESC [ first *)

  val mutable utf8_seq = Buffer.create 10
     (* The current UTF-8 sequence *)

  val mutable utf8_len = 0
     (* Expected length of UTF-8 representation *)			   

  (* --- input side buffers --- *)

  val mutable in_buf = Buffer.create 100

  (* --- dispatch tables --- *)

  val mutable lookup_esc_seq = []

  initializer (
    self # reset();

    lookup_esc_seq <- (
      [ "\027c", true,  self # reset;
        "\027D", true,  self # process_lf;
        "\027E", true,  (fun () -> self # process_cr(); self # process_lf());
        "\027H", true,  (fun () -> tab_stops.(x) <- true);
        "\027M", true,  self # process_rev_lf;
        "\027Z", true,  (fun () -> self # add_input_string "\027[?6c");
        "\0277", true,  self # save_cursor;
        "\0278", true,  self # restore_cursor;
        "\027%@", true, (fun () ->
			   match charset with
			     | Mapping _ -> ()
			     | UTF8 m -> charset <- Mapping m);
        "\027%G", true, (fun () ->
			   match charset with
			     | Mapping m -> charset <- UTF8 m
			     | UTF8 _ -> ());
        "\027%8", true, (fun () ->
			   match charset with
			     | Mapping m -> charset <- UTF8 m
			     | UTF8 _ -> ());
        "\027(B", true, (fun () -> g0_mapping <- ISO);
        "\027(0", true, (fun () -> g0_mapping <- Graphics);
        "\027(U", true, (fun () -> g0_mapping <- CP437);
        "\027)B", true, (fun () -> g1_mapping <- ISO);
        "\027)0", true, (fun () -> g1_mapping <- Graphics);
        "\027)U", true, (fun () -> g1_mapping <- CP437);
        "\027]P", false, (fun () -> if Buffer.length esc_seq = 10 then self # cancel_esc());
          (* ESC ] P: Ignore setting the palette *)
        "\027]R", true, (fun () -> ());
	 (* ESC ] R: Ignore resetting the palette *)

        (* Ignored: ESC # 8, ESC >, ESC =, ESC (U, ESC (K, ESC )U, ESC )K *)

        "\027[[.", true, (fun () -> ());   (* Must be before the following code! *)
        "\027[", false, self # check_csi;
      ] );
  )


  method reset () =
    scroll_first <- 0;
    scroll_last <- height-1;
    tab_stops <- Array.make width false;
    x <- 0;
    y <- 0;
    atts <-  default_atts;
    disp_ctrl <- false;
    toggle_meta <- false;
    insert_mode <- false;
    decckm <- false;
    rel_addr <- false;
    autowrap <- true;
    cur_visible <- true;
    fg_default <- White;
    bg_default <- Black;
    charset <- Mapping ISO;
    g0_mapping <- ISO;
    g1_mapping <- Graphics;

    need_wrap <- false;

    saved_x <- 0;
    saved_y <- 0;
    saved_atts <- default_atts;
    saved_charset <- Mapping ISO;
    saved_g0_mapping <- ISO;
    saved_g1_mapping <- ISO;

    let k = ref 8 in
    while !k < width do
      tab_stops.( !k ) <- true;
      k := !k + 8
    done;

    Buffer.clear esc_seq;
    Buffer.clear utf8_seq;
    Buffer.clear in_buf


  method output_char c =
    (* First check if we are in UTF8 mode. If so, do the UTF8 to Unicode translation
     * first.
     *)
    let uni_c_opt =
      match charset with
	| UTF8 _ ->
	    let cc = Char.code c in
	    if cc <= 0x7f then (
	      Buffer.clear utf8_seq;
	      utf8_len <- 0;
	      Some(cc,cc)
	    )
	    else
	      if cc >= 0x80 && cc <= 0xbf then (
		Buffer.add_char utf8_seq c;
		if Buffer.length utf8_seq = utf8_len then (
		  let s = Buffer.contents utf8_seq in
		  Buffer.clear utf8_seq;
		  utf8_len <- 1;
		  try
		    let ua = Netconversion.uarray_of_ustring `Enc_utf8 s in
		    Some (ua.(0),ua.(0))
		  with
		      _ -> None
		)
		else
		  None
	      )
	      else
		if cc >= 0xc0 && cc <= 0xdf then (
		  Buffer.clear utf8_seq;
		  Buffer.add_char utf8_seq c;
		  utf8_len <- 2;
		  None
		)
		else
		  if cc >= 0xe0 && cc <= 0xef then (
		    Buffer.clear utf8_seq;
		    Buffer.add_char utf8_seq c;
		    utf8_len <- 3;
		    None
		  )
		  else
		    if cc >= 0xf0 && cc <= 0xf7 then (
		      Buffer.clear utf8_seq;
		      Buffer.add_char utf8_seq c;
		      utf8_len <- 4;
		      None
		    )
		    else
		      None  (* ignore *)
	| Mapping m ->
	    let c' = 
	      if toggle_meta then
		Char.chr((Char.code c) lxor 0x80)
	      else
		c in
	    Some(map_char m c', Char.code c)
    in
    match uni_c_opt with
      | None ->
	  (* No character available yet. *)
	  ()
      | Some (mapped_c, uni_c) ->
	  (* Now process the Unicode character uni_c *)
	  (* Special meaning: (-1) = solid block *)
	  let is_utf8 = 
	    match charset with UTF8 _ -> true | _ -> false in
	  ( match uni_c with
	      | 0 -> ()
	      | 7 when is_utf8 || not disp_ctrl -> ()  (* "beep" - ignored *)
	      | 8 -> self # process_bs()
	      | 9 when is_utf8 || not disp_ctrl -> self # process_tab()
	      | 10 -> self # process_lf()
	      | 11 when is_utf8 || not disp_ctrl -> self # process_lf()
	      | 12 -> self # process_lf()
	      | 13 -> self # process_cr()
	      | 14 -> self # process_so()
	      | 15 -> self # process_si()
	      | 24 when is_utf8 || not disp_ctrl -> self # cancel_esc()
	      | 26 -> self # cancel_esc()
	      | 27 -> self # process_esc()
	      | 127 when not disp_ctrl -> ()
	      | 155 -> self # process_csi()
	      | _ when uni_c <= 31 && (is_utf8 || not disp_ctrl) -> ()
	      | _ -> self # process_char mapped_c
	  )

  method private process_esc() =
    Buffer.clear esc_seq;
    Buffer.add_char esc_seq '\027'

  method private process_csi() =
    Buffer.clear esc_seq;
    Buffer.add_char esc_seq '\027';
    Buffer.add_char esc_seq '['

  method private cancel_esc() =
    (* Debug: *)
(*
    eprintf "Last ESC seq:";
    let buf = Buffer.contents esc_seq in
    for k = 0 to String.length buf - 1 do
      let c = Char.code buf.[k] in
      eprintf " %d/%s" c (if c >= 33 && c <= 126 then String.make 1 buf.[k] else "")
    done;
    eprintf "\n";
    flush stderr;
 *)
    (* Real action: *)
    Buffer.clear esc_seq;

  method private check_esc() =
    let is_prefix = ref false in
    let buf = Buffer.contents esc_seq in
    try
      List.iter
	(fun (seq, final, f) ->
	   if final && seq = buf then 
	     (* final: Only match when buf=seq. *)
	     raise (Found (final,f))
	   else
	     if (not final) && String.length seq <= String.length buf &&
	        (seq = String.sub buf 0 (String.length seq)) 
	     then
	       (* Match whenever seq is a prefix of buf *)
	       raise (Found (final,f))
	     else
	       if String.length buf < String.length seq && 
		 (buf = String.sub seq 0 (String.length buf)) 
	       then
		 is_prefix := true
	)
	lookup_esc_seq;
      if not !is_prefix then
	self # cancel_esc()  (* Wrong ESC sequence *)
    with
	Found (final,f) -> 
	  f();
	  if final then self # cancel_esc()

  method private check_csi () =
    let buf = Buffer.contents esc_seq in
    if Str.string_match csi_re buf 0 then (
      (* Found a CSI command! *)
      let qmark = (buf.[2] = '?') in
      let letter = buf.[String.length buf - 1] in
      let params = get_csi_params buf in
      let param0 = List.hd params in  (* exists always *)
      self # cancel_esc();
      ( match (qmark,letter) with
	  | (false, '@') -> self # insert_chars y x param0
	  | (false, 'A') -> self # set_y (y - (max 1 param0))
	  | (false, 'B') -> self # set_y (y + (max 1 param0))
	  | (false, 'C') -> self # set_x (x + (max 1 param0))
	  | (false, 'D') -> self # set_x (x - (max 1 param0))
	  | (false, 'E') -> self # set_y (y + (max 1 param0)); x <- 0
	  | (false, 'F') -> self # set_y (y - (max 1 param0)); x <- 0
	  | (false, 'G') -> self # set_x (param0-1)
	  | (false, 'H') 
	  | (false, 'f') ->
	      ( match params with
		  | row :: col :: _ -> self # set_y ~rel_addr (row-1); self # set_x (col-1)
		  | row :: []       -> self # set_y ~rel_addr (row-1); self # set_x 0
		  | []              -> self # set_y ~rel_addr 0;       self # set_x 0
	      )
	  | (false, 'J') ->
	      ( match param0 with
		  | 0 -> self # erase_line y x (width-1);
		         for k = y+1 to self#max_y do
			   self # erase_line k 0 (width-1);
			 done
		  | 1 -> for k = self#min_y to self#max_y do
		           self # erase_line k 0 (width-1);
		         done
		  | 2 -> self # erase_line y 0 x;
		         for k = self#min_y to y-1 do
			   self # erase_line k 0 (width-1);
			 done
		  | _ -> ()
	      )
	  | (false, 'K') ->
	      ( match param0 with
		  | 0 -> self # erase_line y x (width-1)
		  | 1 -> self # erase_line y 0 x
		  | 2 -> self # erase_line y 0 (width-1)
		  | _ -> ()
	      )
	  | (false, 'L') -> self # insert_lines y (max 1 param0)
	  | (false, 'M') -> self # delete_lines y (max 1 param0)
	  | (false, 'P') -> self # delete_chars y x (max 1 param0)
	  | (false, 'X') -> self # erase_line y x (x + (max 1 param0) - 1)
	  | (false, 'a') -> self # set_x (x + (max 1 param0))
	  | (false, 'c') -> self # add_input_string "\027[?6c"
	  | (false, 'd') -> self # set_y ~rel_addr (param0-1)
	  | (false, 'e') -> self # set_y (y + (max 1 param0))
	  | (false, 'g') ->
	      ( match param0 with
		  | 0 -> tab_stops.(x) <- false
		  | 3 -> Array.fill tab_stops 0 width false
		  | _ -> ()
	      )
	  | (false, 'h') ->
	      ( match param0 with
		  | 3 ->  disp_ctrl <- true
		  | 4 ->  insert_mode <- true
		  | _ ->  ()
	      )
	  | (true,  'h') ->
	      ( match param0 with
		  | 1  ->  decckm <- true
		  | 5  ->  atts <- { atts with reverse = true }
		  | 6  ->  rel_addr <- true
		  | 7  ->  autowrap <- true
		  | 25 ->  cur_visible <- true
		  | _  -> ()
	      )
	  | (false, 'l') ->
	      ( match param0 with
		  | 3 ->  disp_ctrl <- false
		  | 4 ->  insert_mode <- false
		  | _ ->  ()
	      )
	  | (true, 'l') ->
	      ( match param0 with
		  | 1  ->  decckm <- false
		  | 5  ->  atts <- { atts with reverse = false }
		  | 6  ->  rel_addr <- false; 
		           self # set_y scroll_first; 
			  self # set_x 0
		  | 7  ->  autowrap <- false
		  | 25 ->  cur_visible <- false
		  | _  -> ()
	      )
	  | (false, 'm') ->
	      List.iter
		(function
		   | 0 ->    atts <- self # default_atts
		   | 1 ->    atts <- { atts with intense = I_bold }
		   | 2 ->    atts <- { atts with intense = I_dim }
		   | 4 ->    atts <- { atts with underscore = true }
		   | 5 ->    atts <- { atts with blink = true }
		   | 7 ->    atts <- { atts with reverse = true }
		   | 10 ->   charset <- Mapping ISO;   (* Hope this is right *)
			    disp_ctrl <- false;
			    toggle_meta <- false
		   | 11 ->   charset <- Mapping CP437; (* Hope this is right *)
			    disp_ctrl <- true;
			    toggle_meta <- false;
		   | 12 ->   charset <- Mapping CP437; (* Hope this is right *)
			    disp_ctrl <- true;
			    toggle_meta <- true;
		   | 21 ->   atts <- { atts with intense = I_norm }
		   | 22 ->   atts <- { atts with intense = I_norm }
		   | 24 ->   atts <- { atts with underscore = false }
		   | 25 ->   atts <- { atts with blink = false }
		   | 27 ->   atts <- { atts with reverse = false }
		   | 30 ->   atts <- { atts with fg_color = Black }
		   | 31 ->   atts <- { atts with fg_color = Red }
		   | 32 ->   atts <- { atts with fg_color = Green }
		   | 33 ->   atts <- { atts with fg_color = Brown }
		   | 34 ->   atts <- { atts with fg_color = Blue }
		   | 35 ->   atts <- { atts with fg_color = Magenta }
		   | 36 ->   atts <- { atts with fg_color = Cyan }
		   | 37 ->   atts <- { atts with fg_color = White }
		   | 38 ->   atts <- { atts with underscore = true; fg_color = fg_default }
		   | 39 ->   atts <- { atts with underscore = false; fg_color = fg_default}
		   | 40 ->   atts <- { atts with bg_color = Black }
		   | 41 ->   atts <- { atts with bg_color = Red }
		   | 42 ->   atts <- { atts with bg_color = Green }
		   | 43 ->   atts <- { atts with bg_color = Brown }
		   | 44 ->   atts <- { atts with bg_color = Blue }
		   | 45 ->   atts <- { atts with bg_color = Magenta }
		   | 46 ->   atts <- { atts with bg_color = Cyan }
		   | 47 ->   atts <- { atts with bg_color = White }
		   | 49 ->   atts <- { atts with bg_color = bg_default }
		   | _  ->   ()
		)
		params
	  | (false, 'n') ->
	      ( match param0 with
		  | 5 -> self # add_input_string "\027[0n"
		  | 6 -> self # add_input_string (sprintf "\027[%d;%dR" (y+1) (x+1))
		  | _ -> ()
	      )
	  | (false, 'q') -> () (* Set keyboard LEDs... ignored. *)
	  | (false, 'r') ->
	      ( match params with
		  | first :: last :: _ ->
		      let l = if last = 0 then height else last in
		      if first <= l then (
			scroll_first <- max 0 (min (height-1) (first-1));
			scroll_last  <- max 0 (min (height-1) (l-1));
		      )
		  | first :: [] ->
		      scroll_first <- max 0 (min (height-1) (first-1));
		      scroll_last  <- height - 1;
		  | [] ->
		      scroll_first <- 0;
		      scroll_last  <- height - 1;
	      );
	      self # set_y ~rel_addr 0;
	      self # set_x 0
	  | (false, 's') -> self # save_cursor()
	  | (false, 'u') -> self # restore_cursor()
	  | (false, '`') -> self # set_x (param0-1)
	  | (false, ']') ->
	      ( match params with
		  | 8 :: _ ->
		      fg_default <- atts.fg_color;
		      bg_default <- atts.bg_color
		  | _ -> ()
	      )
	  | _ ->
	      (* Unknown command *) 
	      ()
      );
    )

  method private save_cursor() =
    saved_x <- x; saved_y <- y; saved_atts <- atts;
    saved_charset <- charset; saved_g0_mapping <- g0_mapping;
    saved_g1_mapping <- g1_mapping

  method private restore_cursor() =
    x <- saved_x; y <- saved_y; atts <- saved_atts;
    charset <- saved_charset; g0_mapping <- saved_g0_mapping;
    g1_mapping <- saved_g1_mapping;

  method private process_bs() =
    if x > 0 then x <- x - 1;
    need_wrap <- false

  method private process_tab() =
    if x < width - 1 then (
      x <- x + 1;
      while x < width - 1 && not tab_stops.(x) do
	x <- x + 1
      done
    );
    need_wrap <- false

  method private process_cr () =
    x <- 0;
    need_wrap <- false

  method private process_so () =
    match charset with
      | UTF8 _    -> charset <- UTF8 g1_mapping
      | Mapping _ -> charset <- Mapping g1_mapping

  method private process_si () =
    match charset with
      | UTF8 _    -> charset <- UTF8 g0_mapping
      | Mapping _ -> charset <- Mapping g0_mapping

  method private process_lf () =
    if y = scroll_last then
      self # scroll_up()
    else
      self # set_y (y+1);
    need_wrap <- false

  method private process_rev_lf () =
    if y = scroll_first then
      self # scroll_down()
    else
      self # set_y (y-1);
    need_wrap <- false

  method private set_x new_x =
    x <- min (max new_x 0) (width-1);
    need_wrap <- false;

  method private set_y ?(rel_addr = false) new_y =
    let min_y = if rel_addr then scroll_first else 0 in
    let max_y = if rel_addr then scroll_last else height-1 in
    let new_y' = if rel_addr then new_y + scroll_first else new_y in
    y <- max min_y (min max_y new_y');
    need_wrap <- false;

  method private min_y =
    if rel_addr then scroll_first else 0

  method private max_y =
    if rel_addr then scroll_last else height-1

  method private process_char uni_c =
    if Buffer.length esc_seq > 0 then (
      (* Within ESC sequence: *)
      if uni_c <= 127 then (
	Buffer.add_char esc_seq (Char.chr uni_c);
	self # check_esc()
      )
	(* CHECK: else ? *)
    )
    else (
      (* Regular character: *)
      if need_wrap then (
	self # process_cr();
	self # process_lf();
	need_wrap <- false
      );
      if insert_mode then
	self # insert_chars y x 1;
      if uni_c = (-1) then
	screen.(y).(x) <- (Unicode 0xA0, { atts with reverse = not atts.reverse })
      else
	screen.(y).(x) <- (Unicode uni_c, atts);
      if (x < width - 1) then
	x <- x + 1
      else
(*
	if autowrap then (
	  if cur_visible then (
	    self # process_cr();
	    self # process_lf();
	  )
	  else
	    need_wrap <- true
	)
 *)
	need_wrap <- autowrap
    )

  method private default_atts =
    { default_atts with
	fg_color = fg_default;
	bg_color = bg_default
    }

  method private erase_line y_pos x_from x_to =
    (* y_pos, x_from must be valid. x_to can be exaggerated *)
    let l = max 0 (min (width - x_from) (x_to - x_from + 1)) in
    Array.fill screen.(y_pos) x_from l (Empty,atts)

  method private insert_lines y_pos n =
    let max_y = scroll_last in
    let n' = max 0 (min n (max_y - y_pos  + 1)) in
    let n_move = max 0 (max_y - y_pos - n' + 1) in
    Array.blit screen y_pos screen (y_pos+n') n_move;
    for k = y_pos to y_pos + n' - 1 do
      (* Can't use erase_line here, because of cell sharing *)
      screen.(k) <- Array.init width (fun _ -> (Empty, atts))
    done

  method private delete_lines y_pos n =
    let max_y = scroll_last in
    let n' = max 0 (min n (max_y - y_pos + 1)) in
    let n_move = max 0 (max_y - y_pos - n' + 1) in
    Array.blit screen (y_pos+n') screen y_pos n_move;
    for k = y_pos + n_move to y_pos + n_move + n' - 1 do
      (* Can't use erase_line here, because of cell sharing *)
      screen.(k) <- Array.init width (fun _ -> (Empty, atts))
    done

  method private scroll_down() =
    self # insert_lines scroll_first 1

  method private scroll_up() =
    self # delete_lines scroll_first 1

  method private insert_chars y_pos x_pos n =
    let n' = max 0 (min n (width - x_pos)) in
    let n_move = max 0 (width - x_pos - n') in
    Array.blit screen.(y_pos) x_pos screen.(y_pos) (x_pos+n') n_move;
    Array.fill screen.(y_pos) x_pos n' (Empty,atts)

  method private delete_chars y_pos x_pos n =
    let n' = max 0 (min n (width - x_pos)) in
    let n_move = max 0 (width - x_pos - n') in
    Array.blit screen.(y_pos) (x_pos+n') screen.(y_pos) x_pos n_move;
    Array.fill screen.(y_pos) (x_pos+n_move) n' (Empty,atts)

  method add_input_string s =
    Buffer.add_string in_buf s

  method add_input_key (key : key) =
    let s =
      match key with
	| `Translated unicode ->
	    (* Kludge: As we don't support the loadkeys API, we guess the character
	     * codes: For G0/G1 always assume ISO-8859-1, for UTF8 assume UTF-8.
	     *)
	    ( match charset with
		| Mapping _ ->
		    if unicode >= 0 && unicode <= 255 then
		      String.make 1 (Char.chr unicode)
		    else
		      ""
		| UTF8 _ ->
		    ( try
			Netconversion.ustring_of_uchar `Enc_utf8 unicode
		      with
			  Netconversion.Cannot_represent _-> ""
		    )
	    )
	| `Cursor_up -> "\027[A"
	| `Cursor_down -> "\027[B"
	| `Cursor_left -> "\027[D"
	| `Cursor_right -> "\027[C"
	| `Page_up -> "\027[5~"
	| `Page_down -> "\027[6~"
	| `Home -> "\027[1~"
	| `End -> "\027[4~"
	| `Insert -> "\027[2~"
	| `Delete -> "\027[3~"
	| `Backspace -> "\127"
	| `Enter -> "\013"
	| `Tab -> "\009"
	| `Escape -> "\027"
	| `Fn 1 -> "\027[[A"
	| `Fn 2 -> "\027[[B"
	| `Fn 3 -> "\027[[C"
	| `Fn 4 -> "\027[[D"
	| `Fn 5 -> "\027[[E"
	| `Fn 6 -> "\027[17~"
	| `Fn 7 -> "\027[18~"
	| `Fn 8 -> "\027[20~"
	| `Fn 9 -> "\027[21~"
	| `Fn 10 -> "\027[22~"
	| `Fn 11 -> "\027[23~"
	| `Fn 12 -> "\027[24~"
	| `Fn _ -> ""
    in
    self # add_input_string s

  method input_string() =
    let s = Buffer.contents in_buf in
    Buffer.clear in_buf;
    s

  method screen = screen

  method screen_as_html =
    let b = Buffer.create 1024 in
    for y_pos = 0 to height - 1 do
      let last_atts = ref None in
      for x_pos = 0 to width - 1 do
	let (data, atts) = screen.(y_pos).(x_pos) in
	let atts =
	  if data = Empty then
	    { atts with intense = I_norm; underscore = false; reverse = false } 
	  else
	    atts in
	if !last_atts <> Some atts then (
	  ( match !last_atts with
	      | None -> ()
	      | Some _ -> 
		  Buffer.add_string b "</span>"
	  );
	  let classes =
	    ( match atts.intense with
		| I_norm -> []
		| I_dim  -> [ "t-dim" ]
		| I_bold -> [ "t-bold" ] ) @
	    ( if atts.underscore then [ "t-underscore" ] else [] ) @
	    ( if atts.blink then [ "t-blink" ] else [] ) @
	    ( if atts.reverse then 
		[ "t-fg-" ^ color atts.bg_color; "t-bg-" ^ color atts.fg_color ]
	      else
		[ "t-fg-" ^ color atts.fg_color; "t-bg-" ^ color atts.bg_color ] ) in
	  Buffer.add_string b (sprintf "<span class='%s'>" (String.concat " " classes));
	);
	last_atts := Some atts;
	let is_cursor_pos = x_pos = x && y_pos = y in
	if is_cursor_pos then
	  Buffer.add_string b "<span id='t-cursor'>";
	( match data with
	      Empty -> Buffer.add_string b "&nbsp;"
	    | Unicode 32 -> Buffer.add_string b "&nbsp;"
	    | Unicode n -> 
		if (n > 32 && n <= 127) || n >= 160 then
		  Buffer.add_string b ("&#" ^ string_of_int n ^ ";")
	);
	if is_cursor_pos then
	  Buffer.add_string b "</span>";
      done;
      Buffer.add_string b "</span><br>"
    done;
    Buffer.contents b

  method width = width

  method height = height

  method cursor_x = x

  method cursor_y = y

  method cursor_visible = cur_visible

end

