(*
 * <COPYRIGHT>
 * Copyright 2005 Gerd Stolpmann
 *
 * <GPL>
 * This file is part of UMLMON.
 *
 * UMLMON is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * UMLMON is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with UMLMON; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * </>
 *)

(* $Id: umlmon_shared.ml 209 2005-10-23 13:46:46Z gerd $ *)

open Umlmon_types
open Umlmon_common
open Printf

let setup_shared global =
  (* Important note: this function may be called concurrently by several
   * umlmon instances. Creating the directories is not problematic; 
   * either they exist or they don't exist. Same holds for the device
   * nodes.
   *)
  let shared = global.umlshared in
  (* First create the directories: *)
  List.iter
    (fun (dir, mode, uid, gid) -> 
       try
	 Unix.mkdir dir mode;    (* modulo umask = 0o022 *)
	 Unix.chown dir uid gid;
       with
	 | Unix.Unix_error(Unix.EEXIST,_,_) -> ()
    )
    [ shared, 0o755, 0, 0;
      (Filename.concat shared "disks"), 0o755, 0, 0;
      (Filename.concat shared "kernels"), 0o755, 0, 0;
    ];
  (* Create "locks" and "macpool": *)
  List.iter
    (fun name ->
       try
	 let fd = 
	   Unix.openfile (Filename.concat shared name)
	     [ Unix.O_WRONLY; Unix.O_CREAT; Unix.O_EXCL ]
	     0o600 in
	 Unix.close fd
       with
	 | Unix.Unix_error(Unix.EEXIST,_,_) -> ()
    )
    [ "locks"; "macpool" ];
  (* Create the missing device nodes, and/or check them: *)
  List.iter
    (fun dev ->
       let s = Unix.LargeFile.stat dev.dev_target in
       let p = Filename.concat 
	 (Filename.concat shared "disks") dev.dev_name in
       let mknod_type =
	 match s.Unix.LargeFile.st_kind with
	   | Unix.S_CHR -> Linux.File.S_IFCHR s.Unix.LargeFile.st_rdev
	   | Unix.S_BLK -> Linux.File.S_IFBLK s.Unix.LargeFile.st_rdev
	   | _ ->
	       failwith("Bad file type: " ^ dev.dev_target)
       in
       try
	 Linux.File.mknod p dev.dev_create_perm mknod_type;
	 Unix.chmod p dev.dev_create_perm;
	 Unix.chown p 0 dev.dev_create_group
       with
	 | Unix.Unix_error(Unix.EEXIST,_,_) ->
	     (* Maybe we have to check the nodes: *)
	     ( match dev.dev_identity with
		 | `Ignore -> ()
		 | `Check | `Update ->
		     let t = Unix.LargeFile.stat p in
		     let mismatch =
		       t.Unix.LargeFile.st_kind <> s.Unix.LargeFile.st_kind
		       || t.Unix.LargeFile.st_rdev <> s.Unix.LargeFile.st_rdev
		     in
		     if mismatch then (
		       if dev.dev_identity = `Check then
			 failwith("Device identity mismatch: " ^ p)
		       else (  (* Update; keep permissions! *)
			 Unix.unlink p;
			 Linux.File.mknod p t.Unix.LargeFile.st_perm mknod_type;
			 Unix.chmod p t.Unix.LargeFile.st_perm;
			 Unix.chown p t.Unix.LargeFile.st_uid t.Unix.LargeFile.st_gid;
		       )
		     )
	     )
    )
    global.umldevs;
;;

	    
type lock = [ `Read | `Write ]

(* Lock file format:
 * - First line contains the next free lock_id
 * - Every following line describes a lock
 * - Line has format
 *   <file>:<user>:<type>:<lock_id>
 * - <file> is the jail name of the file (e.g. /shared/disks/foo)
 * - <vm> is usually "vm/" ^ umlid
 * - <type> is either "R" for a read lock or "W" for a write lock.
 * - Special encodings:
 *   %3a for ":"
 *   %25 for "%"
 *)


type entry = jailfile * string * lock * lock_id
   (* An entry of the lock file *)

let colon_re = Str.regexp ":";;
let percent_decode_re = Str.regexp "%[0-9a-fA-F][0-9a-fA-F]";;
let percent_encode_re = Str.regexp "[%:]";;

let percent_decode s =
  Str.global_substitute
    percent_decode_re
    (fun s ->
       let hex = String.sub (Str.matched_string s) 1 2 in
       let n = int_of_string ("0x" ^ hex) in
       String.make 1 (Char.chr n))
    s
;;


let percent_encode s =
  Str.global_substitute
    percent_encode_re 
    (fun s ->
       match Str.matched_string s with
	 | "%" -> "%25"
	 | ":" -> "%3a"
	 | _   -> assert false)
    s
;;


let lock_decode s =
  match s with
    | "R" -> `Read
    | "W" -> `Write
    | _   -> failwith "lock_decode"
;;


let lock_encode l =
  match l with
    | `Read -> "R"
    | `Write -> "W"
;;


let read_lock_file fd =
  let f = Unix.in_channel_of_descr fd in
  let next_id = 
    try
      int_of_string(input_line f)
    with End_of_file -> 0
  in
  let entries = ref [] in
  try
    while true do
      let line = input_line f in
      let fields = Str.split_delim colon_re line in
      match fields with
	| [ file; user; lock_t; lock_id_s ] ->
	    let file = percent_decode file in
	    let user = percent_decode user in
	    let lock_t = lock_decode lock_t in
	    let lock_id = Lock_id (int_of_string lock_id_s) in
	    let entry = (Jailfile file, user, lock_t, lock_id) in
	    entries := entry :: !entries
	| _ ->
	    failwith ("Bad line in lock file: " ^ line)
    done;
    assert false
  with
    | End_of_file ->
	(next_id, List.rev !entries)
;;


let write_lock_file fd next_id entries =
  ignore(Unix.lseek fd 0 Unix.SEEK_SET);
  Unix.ftruncate fd 0;
  let f = Unix.out_channel_of_descr fd in
  fprintf f "%d\n" next_id;
  List.iter
    (fun (Jailfile file, user, lock, Lock_id lock_id) ->
       let file = percent_encode file in
       let user = percent_encode user in
       let lock = lock_encode lock in
       fprintf f "%s:%s:%s:%d\n" file user lock lock_id
    )
    entries;
  flush f
;;


let lock_transaction global f =
  let lock_file = Filename.concat global.umlshared "locks" in
  let fd = 
    Unix.openfile lock_file [ Unix.O_RDWR ] 0 in
  try
    restart (Unix.lockf fd Unix.F_LOCK) 1;
    let next_id, entries = read_lock_file fd in
    let r, next_id', entries' = f next_id entries in
    write_lock_file fd next_id' entries';
    Unix.close fd;  (* Releases all advisory locks *)
    r
  with
    | error ->
	Unix.close fd; raise error
;;


exception Return_locks of (lock * jailfile * bool) list ;;


let get_locks global spec =
  let is_our_lock user =
    user = "vm/" ^ spec.umlid in

  try
    lock_transaction
      global
      (fun _ entries ->
	 let l =
	   List.map
	     (fun (file, user, lock, _) ->
		(lock, file, is_our_lock user))
	     entries in
	 raise (Return_locks l)
      );
    assert false
  with
    | Return_locks l -> l
;;


let lock_disks global spec (locks : (lock * jailfile) list) =
  (* Check that [locks] is non-contradictory: *)
  List.iter
    (fun (l, file) ->
       if List.exists (fun (l', file') -> file = file' && l <> l') locks then
	 let Jailfile f = file in
	 raise (Lock_failure f)
    )
    locks;

  let user = "vm/" ^ spec.umlid in

  lock_transaction
    global
    (fun next_id entries ->
       (* Check whether there are already locks held by other VMs: *)
       List.iter
	 (fun (l, Jailfile file) ->
	    ( match l with
		| `Read ->
		    if List.exists 
		         (fun (Jailfile jf, id, l, _) ->
			    jf = file && l = `Write) entries
		    then raise(Lock_failure file)
		| `Write ->
		    if List.exists 
		         (fun (Jailfile jf, id, _, _) ->
			    jf = file) entries
		    then raise(Lock_failure file)
	    )
	 )
	 locks;
       let new_entries =
	 List.map
	   (fun (l, f) -> (f, user, l, Lock_id next_id))
	   locks in
       (Lock_id next_id, next_id+1, new_entries @ entries)
    )
;;


let unlock_disks global spec lock_id =
  lock_transaction
    global
    (fun next_id entries ->
       let entries' =
	 List.filter
	   (fun (f, _, _, lid) ->
	      lid <> lock_id)
	   entries in
       ( (), next_id, entries' )
    )
;;


let unlock_all_disks global spec =
  let user = "vm/" ^ spec.umlid in
  lock_transaction
    global
    (fun next_id entries ->
       let entries' =
	 List.filter
	   (fun (Jailfile jf, id, _, _) ->
	      id <> user)
	   entries in
       ( (), next_id, entries' )
    )
;;


(* MAC pool file format:
 * - Every line is an entry and has format:
 *   <hwaddr>:<ipaddr>:<vm>
 * - <hwaddr> is the :-separated MAC address
 * - <ipaddr> is the .-separated IP address (optional)
 * - <vm> is "" when unsused, or "vm/" ^ identifier
 * - all fields are percent-encoded (see above)
 *)

type macpool_entry = hwaddr * string option * string option

let read_macpool_file fd =
  let f = Unix.in_channel_of_descr fd in
  let entries = ref [] in
  try
    while true do
      let line = input_line f in
      let fields = Str.split_delim colon_re line in
      match fields with
	| [ hwaddr_s; ipaddr; vm ] ->
	    let hwaddr = parse_hwaddr(percent_decode hwaddr_s) in
	    let ipaddr = 
	      if ipaddr = "" then None else Some(percent_decode ipaddr) in
	    let vm = if vm = "" then None else Some (percent_decode vm) in
	    let entry = (hwaddr, ipaddr, vm) in
	    entries := entry :: !entries
	| _ ->
	    failwith ("Bad line in macpool file: " ^ line)
    done;
    assert false
  with
    | End_of_file ->
	List.rev !entries
;;


let write_macpool_file fd entries =
  ignore(Unix.lseek fd 0 Unix.SEEK_SET);
  Unix.ftruncate fd 0;
  let f = Unix.out_channel_of_descr fd in
  List.iter
    (fun (hwaddr, ipaddr, vm_opt) ->
       let hwaddr = percent_encode(format_hwaddr hwaddr) in
       let ipaddr = match ipaddr with None -> "" | Some s -> percent_encode s in
       let vm = match vm_opt with None -> "" | Some s -> percent_encode s in
       fprintf f "%s:%s:%s\n" hwaddr ipaddr vm
    )
    entries;
  flush f
;;


let macpool_transaction global f =
  let macpool_file = Filename.concat global.umlshared "macpool" in
  let fd = 
    Unix.openfile macpool_file [ Unix.O_RDWR ] 0 in
  try
    restart (Unix.lockf fd Unix.F_LOCK) 1;
    let entries = read_macpool_file fd in
    let r, entries' = f entries in
    write_macpool_file fd entries';
    Unix.close fd;  (* Releases all advisory locks *)
    r
  with
    | error ->
	Unix.close fd; raise error
;;


let reserve_hwaddr global spec ip_opt = 
  let user = "vm/" ^ spec.umlid in

  let new_addr pred =
    (* Find a new address that satisfies pred *)
    let rec search_new_addr addr n =
      if n = 0 then 
	raise Not_found
      else
	if pred addr then
	  addr
	else
	  let addr' = increment_hwaddr addr in
	  search_new_addr addr' (n-1)
    in
    let (addr,n) = global.umlhwpool in
    search_new_addr addr n
  in

  let reuse_addr addr entries =
    List.map
      (fun (h, ip_opt', vm_opt') -> 
	 if h = addr then
	   (h, ip_opt, Some user)
	 else
	   (h, ip_opt', vm_opt')
      )
      entries
  in

  macpool_transaction
    global
    (fun entries ->
       (* If we have an IP address, first check whether there is a 
        * corresponding free MAC address.
        *)
       ( try
	   ( match ip_opt with
	       | None -> raise Not_found
	       | Some ip ->
		   let (my_hwaddr, _, _) =
		     List.find
		       (fun (_, ip_opt', vm_opt') ->
			  ip_opt' = ip_opt && vm_opt' = None)
		       entries 
		       (* or Not_found *)
		   in
		   let entries' = reuse_addr my_hwaddr entries in
		   (my_hwaddr, entries')
	   )
	 with Not_found ->
	   (* Otherwise, we need to allocate a new address. First check
            * whether there is a free MAC address without IP address.
            *)
	   ( try
	       let (my_hwaddr, _, _) =
		 List.find
		   (fun (_, ip_opt', vm_opt) ->
		      ip_opt' = None && vm_opt = None)
		   entries 
		   (* or Not_found *)
	       in
	       let entries' = reuse_addr my_hwaddr entries in
	       (my_hwaddr, entries')
	     with
		 Not_found ->
		   (* Otherwise, get a totally unused address *)
		   ( try
		       let my_hwaddr = 
			 new_addr
			   (fun addr ->   (* express "unused address" *)
			      not(List.exists
				    (fun (h, _, _) -> h = addr)
				    entries)) in
		       let entries' =
			 (my_hwaddr, ip_opt, Some user) :: entries in
		       (my_hwaddr, entries')
		     with
			 Not_found ->
			   (* Last chance: Get any free address *)
			   ( try
			       let (my_hwaddr, _, _) =
				 List.find
				   (fun (_, ip_opt', vm_opt) -> vm_opt = None)
				   entries 
				   (* or Not_found *)
			       in
			       let entries' = reuse_addr my_hwaddr entries in
			       (my_hwaddr, entries')
			     with
				 Not_found ->
				   failwith "Out of pool MAC addresses"
			   )
		   )
	   )
       )    
    )
;;	     


let release_all_hwaddr global spec =
  let user = "vm/" ^ spec.umlid in
  macpool_transaction
    global
    (fun entries ->
       let entries' =
	 List.map
	   (fun (h, ip_opt, vm_opt) -> 
	      if vm_opt = Some user then
		(h, ip_opt, None)
	      else
		(h, ip_opt, vm_opt)
	   ) 
	   entries in
       ( (), entries' )
    )
;;


let get_kernels global =
  let kernels_dir = Filename.concat global.umlshared "kernels" in
  if not (Sys.file_exists kernels_dir) then
    []
  else (
    let files = Array.to_list (Sys.readdir kernels_dir) in
    let files' =
      List.filter
	(fun file ->
	   let abs_file = Filename.concat kernels_dir file in
	   let s = Unix.LargeFile.stat abs_file in
	   s.Unix.LargeFile.st_perm land 1 <> 0
	)
	files in
    List.map
      (fun file -> Jailfile ("/shared/kernels/" ^ file))
      files'
  )
;;
