(*
 * <COPYRIGHT>
 * Copyright 2005 Gerd Stolpmann
 *
 * <GPL>
 * This file is part of UMLMON.
 *
 * UMLMON is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * UMLMON is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with UMLMON; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * </>
 *)

(* $Id: umlmon_term.ml 255 2007-07-01 15:31:22Z gerd $ *)

open Umlmon_types
open Umlmon_common
open Umlmonctrl_aux
open Termemu


let find_term umlvm name : umlterm_activation =
  let actv =
    match umlvm with
      | `Inactive -> raise Not_found
      | `Finishing -> raise Not_found
      | `Stopped a -> a
      | `Running a -> a
      | `Starting a -> a in
  List.find (fun a -> a.actv_termname = name) actv.actv_umlterms
;;


let unconnected_term ptyinch =
  let buf = String.create 8192 in
  let enable_drop_data = ref true in
  let drop_data() =
    try
      if !enable_drop_data then (
	while ptyinch # input buf 0 (String.length buf) > 0 do () done
      );
      !enable_drop_data
    with
	End_of_file -> 
	  false
  in
  let stop_drop_data() =
    enable_drop_data := false
  in
  
  ptyinch # request_notification drop_data;
  `Unconnected stop_drop_data
;;

let leave_ptystate =
  function
    | `Unconnected f -> f()
    | `Connected _ -> ()
    | `Connected_with _ -> ()

  

let close_term_fds actv_terms =
  (* Closes the descriptors of all terminals *)
  List.iter
    (fun at ->
       match at.actv_termstatus with
	 | `Pty p ->
	     Unix.close p.ptymaster;
	     ( match p.ptylog with
		 | None -> ()
		 | Some fd -> Unix.close fd;

	     )
    )
    actv_terms

let stop_term_engines actv_terms =
  (* Stop the engines of all terminals *)
  List.iter
    (fun at ->
       match at.actv_termstatus with
	 | `Pty p ->
	     p.ptyinch_eng # abort();
	     p.ptyoutch_eng # abort();
    )
    actv_terms


let shutdown_terminals actv_terms =
  close_term_fds actv_terms;
  stop_term_engines actv_terms
;;

let reset_connect_timer ues pty =
  match pty.ptytimer with
    | None -> ()
    | Some g -> 
	Unixqueue.clear ues g;
	pty.ptytimer <- None
  

let do_disconnect ues at pty =
  reset_connect_timer ues pty;
  pty.ptytimer <- None;
  ( match pty.ptystate with
      | `Connected _ ->
	  pty.ptystate <- unconnected_term pty.ptyinch;
	  pty.ptyinch # notify();   (* Wake proc_read/proc_write up if necessary *)
	  pty.ptyoutch # notify();
	  at.actv_termemu <- None;
	  monitor_log ("umlmon: Terminal is disconnected: " ^ 
			 at.actv_termname)
      | `Connected_with(_, rd_eng_opt, wr_eng_opt) ->
	  ( match rd_eng_opt with
	      | Some rd_eng -> rd_eng # abort()
	      | None -> ()
	  );
	  ( match wr_eng_opt with
	      | Some wr_eng -> wr_eng # abort()
	      | None -> ()
	  );
	  pty.ptystate <- unconnected_term pty.ptyinch;
	  pty.ptyinch # notify();
	  pty.ptyoutch # notify();
	  at.actv_termemu <- None;
	  monitor_log ("umlmon: Terminal is disconnected: " ^ 
			 at.actv_termname)
      | _ -> ()
  )
;;


let disconnect_terminals ues actv_terms =
  List.iter
    ( fun at ->
	match at.actv_termstatus with
	  | `Pty pty ->
	      do_disconnect ues at pty
    )
    actv_terms


let rotate_terminals spec actv_terms =
  List.iter
    (fun at ->
       match at.actv_termstatus with
	 | `Pty p ->
	     ( match p.ptylog with
		 | None -> ()
		 | Some fd ->
		     ( match at.actv_termdef with
			 | `Pty(Some name) ->
			     let fd' = open_log_file spec name in
			     Unix.dup2 fd' fd;
			     Unix.close fd'
			 | _ -> ()
		     )
	     )
    )
    actv_terms

(**********************************************************************)
(* Buffer classes *)


(* The following class is a buffer that has both an input and an output interface.
 * The input interface complies to async_in_channel, and the output interface
 * complies to async_out_channel.
 *
 * If [copy_fd] is passed, all output data are additionally written to this file.
 *)

exception Broken_pipe

class term_buffer ?(get_copy_fd = fun () -> None) () =
object(self)
  val mutable notify_list = []
  val mutable notify_list_new = []

  val buf = String.create 8192
  val mutable buflen = 0
  val mutable eof = false
  val mutable pos_in = 0
  val mutable pos_out = 0
  val mutable closed_in = false
  val mutable closed_out = false

  method input s spos slen =
    if closed_in then raise Netchannels.Closed_channel;
    if spos < 0 || slen < 0 || spos + slen > String.length s then
      invalid_arg "term_input#input";
    let old_can_input = self # can_input in
    let old_can_output = self # can_output in
    let len = min slen buflen in
    if len = 0 && slen > 0 && eof then raise End_of_file;
    String.blit buf 0 s spos len;
    String.blit buf len buf 0 (buflen - len);
    buflen <- buflen - len;
    pos_in <- pos_in + len;
    if old_can_input <> self # can_input || old_can_output <> self # can_output then
      self # notify();
    len

  method can_input =
    eof || buflen > 0

  method pos_in = 
    pos_in

  method output s spos slen =
    if closed_out then raise Netchannels.Closed_channel;
    if closed_in then raise Broken_pipe;
    if spos < 0 || slen < 0 || spos + slen > String.length s then
      invalid_arg "term_output#output";
    let old_can_input = self # can_input in
    let old_can_output = self # can_output in
    let space = String.length buf - buflen in
    let len = min space slen in
    String.blit s spos buf buflen len;
    buflen <- buflen + len;
    pos_out <- pos_out + len;
    ( try
	( match get_copy_fd() with
	    | Some fd ->
		ignore(Unix.write fd s spos len)
	    | None ->
		()
	)
      with
	| Unix.Unix_error(_, _, _) ->
	    monitor_log ("umlmon: Cannot write to log file");
    );
    if old_can_input <> self # can_input || old_can_output <> self # can_output then
      self # notify();
    len

  method can_output =
    buflen < String.length buf && not eof

  method pos_out = 
    pos_out

  method close_in() =
    if closed_in then raise Netchannels.Closed_channel;
    closed_in <- true

  method close_out() =
    if closed_out then raise Netchannels.Closed_channel;
    eof <- true;
    closed_out <- true;
    self # notify();   (* buffer no longer empty *)

  method request_notification f =
    notify_list_new <- f :: notify_list_new

  method notify() =
    notify_list <- notify_list @ notify_list_new;
    notify_list_new <- [];
    notify_list <- List.filter (fun f -> f()) notify_list

  method flush () =
    ()

end

(**********************************************************************)
(* Terminal setup *)

let create_terminals ues umlspec umlvm =
  let uid = umlspec.umlpriv.priv_uid in
  let gid = umlspec.umlpriv.priv_gid in

  let fd_list = ref [] in

  try
    let actv_terms =
      List.map
	(fun ut ->
	   let s = 
	     match ut.termdef with
	       | `Pty log_opt ->
		   let fd_master = Unix.openfile "/dev/ptmx" [Unix.O_RDWR] 0 in
		   fd_list := fd_master :: !fd_list;
		   let name_slave = Linux.Pty.ptsname fd_master in
		   (* After calling grantpt, do chmod/chown directly.
		    * We are root, so we are allowed to do so.
		    *)
		   Linux.Pty.grantpt fd_master;
		   let st = Unix.LargeFile.stat name_slave in
		   Unix.chown name_slave uid gid;
		   Linux.Pty.unlockpt fd_master;
(* -- We must not open+close the slave here! Otherwise the master sees immediately EIO!
		   let fd_slave = 
		     Unix.openfile name_slave [Unix.O_RDWR; Unix.O_NOCTTY ] 0 in
		   Unix.close fd_slave;
 *)
		   let fd_log =
		     match log_opt with
		       | Some logfile ->
			   let fd = open_log_file umlspec logfile in
			   fd_list := fd :: !fd_list;
			   fd_log ~nl:true fd "umlmon: Startup";
			   Some fd;
		       | None -> None
		   in

		   let get_copy_fd = 
		     match fd_log with
		       | Some fd -> 
			   ( fun () -> Some fd )
		       | None -> (fun () -> None) in
		   let ptyinch = 
		     new term_buffer ~get_copy_fd () in

		   let ptyinch_fd = Unix.dup fd_master in
		   Unix.set_nonblock ptyinch_fd;

		   let ptyinch_eng =
		     new Uq_engines.receiver 
		       ~src:ptyinch_fd
		       ~dst:ptyinch 
		       ~close_src:true
		       ~close_dst:true
		       ues in
		   Uq_engines.when_state
		     ~is_done:(fun _ ->
				 monitor_log("umlmon: EOF on terminal " ^ 
					       ut.termname)
				 (* Should not happen... An EIO error is expected
				  * when the client closes the terminal device.
				  *)
			      )
		     ~is_aborted:(fun _ ->
				 monitor_log("umlmon: Deactivated terminal " ^ 
					       ut.termname)
			      )
		     ~is_error:(function
				  | Unix.Unix_error(Unix.EIO,_,_) ->
				      monitor_log("umlmon: Terminal closed: " ^ 
						    ut.termname)
				  | err ->
				      monitor_log("umlmon: Error for terminal " ^ 
						    ut.termname ^ ": " ^ 
						    Printexc.to_string err)
			       )
		     ptyinch_eng;

		   let ptyoutch =
		     new term_buffer () in

		   let ptyoutch_fd = Unix.dup fd_master in
		   Unix.set_nonblock ptyoutch_fd;

		   let ptyoutch_eng =
		     new Uq_engines.sender
		       ~src:ptyoutch
		       ~dst:ptyoutch_fd
		       ~close_src:false
		       ~close_dst:true
		       ues in

		   let pty =
		     { ptymaster = fd_master;
		       ptyslavename = name_slave;
		       ptylog = fd_log;
		       ptylogname = log_opt;
		       ptyinch = (ptyinch :> pty_in_channel);
		       ptyinch_eng = ptyinch_eng;
		       ptyoutch = (ptyoutch :> pty_out_channel);
		       ptyoutch_eng = ptyoutch_eng;
		       ptystate = unconnected_term ptyinch;
		       ptytimer = None;
		     } in
		   `Pty pty
	   in
	   
	   { actv_termname = ut.termname;
	     actv_termdef = ut.termdef;
	     actv_termstatus = s;
	     actv_termemu = None
	   }
	)
	umlspec.umlterms in
    actv_terms
      
  with err ->
    (* Close all files before re-raising [err]: *)
    List.iter Unix.close !fd_list;
    raise err
;;


let set_connect_timer ues name ct_id at pty =
  reset_connect_timer ues pty;
  let g = Unixqueue.new_group ues in
  pty.ptytimer <- Some g;
  Unixqueue.once ues g 60.0 (fun () ->
			       match pty.ptystate with
				 | `Connected id ->
				     if ct_id = id then
				       do_disconnect ues at pty
				 | _ -> ())
;;


let next_conn_id = ref 0 ;;

let proc_connect ues gs =
  protect
    (fun session name reply ->
       Umlmon_auth.ensure_authenticated session;
       let umlvm = gs.umlvm in
       try
	 let at = find_term umlvm name in  (* or Not_found *)
	 let conn_id = Int32.of_int !next_conn_id in
	 incr next_conn_id;
	 ( match at.actv_termstatus with
	     | `Pty pty ->
		 ( match pty.ptystate with
		     | `Connected _
		     | `Connected_with _  -> 
			 failwith "Terminal is already connected"
		     | _ -> ()
		 );
		 leave_ptystate pty.ptystate;
		 pty.ptystate <- `Connected conn_id;
		 set_connect_timer ues name conn_id at pty;
		 monitor_log ("umlmon: Terminal is connected: " ^ name)
	 );
	 let handle =
	   { ct_name = name;
	     ct_id = conn_id } in
	 reply (`code_ok handle)
       with
	 | Not_found ->
	     reply (`code_error "Terminal not found")
	 | Failure msg ->
	     reply (`code_error msg)
    )
;;


let drop_root umlspec f arg =
  (* Execute f arg without having euid=0 privileges *)
  let back_to_root() =
    Linux.Privilege.setreuid 0 0;
    Linux.Privilege.setregid 0 0
      (* initgroups? not needed because root is superuser anyway *)
  in
  let uid = umlspec.umlpriv.priv_uid in
  let gid = umlspec.umlpriv.priv_gid in
  Linux.Privilege.initgroups umlspec.umlvmuser gid;
  Linux.Privilege.setregid 0 gid;
  Linux.Privilege.setreuid 0 uid;
  try
    let res = f arg in
    back_to_root();
    res
  with error ->
    back_to_root();
    raise error
;;


let open_file_or_connect filename rd_flag wr_flag =
  let st = Unix.stat filename in
  if st.Unix.st_kind = Unix.S_SOCK then
    let fd = Unix.socket Unix.PF_UNIX Unix.SOCK_STREAM 0 in
    ( try
	Unix.connect fd (Unix.ADDR_UNIX filename);
	fd
      with
	| e -> Unix.close fd; raise e
    )
  else
    Unix.openfile filename
      (match rd_flag, wr_flag with
	 | true, true -> [ Unix.O_RDWR ]
	 | true, false -> [ Unix.O_RDONLY ]
	 | false, true -> [ Unix.O_WRONLY ]
	 | _ -> assert false
      )
      0
;;


let proc_connect_with ues gs =
  protect
    (fun session (name, host_file, rd_flag, wr_flag) reply ->
       Umlmon_auth.ensure_authenticated session;
       let umlvm = gs.umlvm in
       let umlspec = Lazy.force gs.umlspec in
       try
	 let at = find_term umlvm name in  (* or Not_found *)
	 let conn_id = Int32.of_int !next_conn_id in
	 incr next_conn_id;
	 ( match at.actv_termstatus with
	     | `Pty pty ->
		 ( match pty.ptystate with
		     | `Connected _ 
		     | `Connected_with _ -> 
			 failwith "Terminal is already connected"
		     | _ -> ()
		 );
		 if not rd_flag && not wr_flag then
		   failwith "connect_with: invalid arguments";
		 let fd = 
		   drop_root
		     umlspec
		     (open_file_or_connect host_file rd_flag)
		     wr_flag in

		 let rd_eng_opt =
		   if rd_flag then
		     let rd_fd = Unix.dup fd in
		     Unix.set_nonblock rd_fd;
		     let rd_eng =
		       new Uq_engines.receiver 
		       ~src:rd_fd
		       ~dst:pty.ptyoutch 
		       ~close_src:true
		       ~close_dst:false
		       ues in
		     Some rd_eng
		   else
		     None in
		 
		 let wr_eng_opt =
		   if wr_flag then
		     let wr_fd = Unix.dup fd in
		     Unix.set_nonblock wr_fd;
		     let wr_eng =
		       new Uq_engines.sender
		       ~src:pty.ptyinch
		       ~dst:wr_fd
		       ~close_src:false
		       ~close_dst:true
		       ues in
		     Some wr_eng
		   else
		     None in

		 Unix.close fd;

		 leave_ptystate pty.ptystate;
		 reset_connect_timer ues pty;
		 
		 pty.ptystate <- `Connected_with(conn_id, rd_eng_opt, wr_eng_opt);
		 monitor_log ("umlmon: Terminal is connected: " ^ name)
	 );
	 let handle =
	   { ct_name = name;
	     ct_id = conn_id } in
	 reply (`code_ok handle)
       with
	 | Not_found ->
	     reply (`code_error "Terminal not found")
	 | Failure msg ->
	     reply (`code_error msg)
    )
;;

let proc_read ues gs session (handle,timeout) reply =
  (* proc_read is asynchronous and cannot be protected against exceptions! *)
  let name = handle.ct_name in
  let umlvm = gs.umlvm in
  let buf = String.create 8192 in
  let t0 = Unix.gettimeofday() in
  let g = Unixqueue.new_group ues in
  let do_input pty () =
    let do_cont = ref false in
    protect 
      (fun session () reply ->
	 try
	   ( match pty.ptystate with
	       | `Connected id -> if id <> handle.ct_id then raise End_of_file
	       | _ -> raise End_of_file
	   );
	   if Unix.gettimeofday() >= t0 +. Int32.to_float timeout then raise End_of_file;
	   let n = pty.ptyinch # input buf 0 (String.length buf) in
	   if n > 0 then (
	     Unixqueue.clear ues g;  (* Stop timer *)
	     reply (`code_ok (String.sub buf 0 n));
	   )
	   else 
	     do_cont := true
	 with
	   | End_of_file ->
	       Unixqueue.clear ues g;  (* Stop timer *)
	       reply `code_interrupt ;
      )
      session () reply;
    !do_cont
  in

  protect
    (fun session () reply ->
       try
	 Umlmon_auth.ensure_authenticated session;
	 let at = find_term umlvm name in  (* or Not_found *)
	 ( match at.actv_termstatus with
	     | `Pty pty ->
		 ( match pty.ptystate with
		     | `Connected id -> if id <> handle.ct_id then raise Not_found
		     | _ -> raise Not_found
		 );
		 set_connect_timer ues name handle.ct_id at pty;
		 if do_input pty () then (
		   pty.ptyinch # request_notification (do_input pty);
		   Unixqueue.once ues g (Int32.to_float timeout +. 0.1) pty.ptyinch#notify;
		 )
	 )
       with
	 | Not_found ->
	     reply (`code_error "Terminal not found or not connected")
    )
    session () reply
;;

let proc_write ues gs session (handle,data,timeout) reply =
  (* proc_write is asynchronous and cannot be protected against exceptions! *)
  let name = handle.ct_name in
  let umlvm = gs.umlvm in
  let pos = ref 0 in
  let t0 = Unix.gettimeofday() in
  let g = Unixqueue.new_group ues in
  let do_output pty () =
    let do_cont = ref false in
    protect 
      (fun session () reply ->
	 let may_output =
	   match pty.ptystate with
	     | `Connected id -> 
		 id = handle.ct_id  &&  
		      Unix.gettimeofday() < t0 +. Int32.to_float timeout 
	     | _ -> false
	 in

	 if may_output then (
	   let l = String.length data - !pos in
	   let n = pty.ptyoutch # output data !pos l in
	   pos := !pos + n;
	   if n < l then (
	     do_cont := true    (* keep this function enabled *)
	   )
	   else (
	     Unixqueue.clear ues g;  (* Stop timer *)
	     reply `code_ok;
	   )
	 )
	 else (
	   Unixqueue.clear ues g;  (* Stop timer *)
	   reply (`code_interrupt (Int32.of_int !pos));
	 )
      )
      session () reply;
    !do_cont
  in


  protect
    (fun session () reply ->
       try
	 Umlmon_auth.ensure_authenticated session;
	 let at = find_term umlvm name in  (* or Not_found *)
	 ( match at.actv_termstatus with
	     | `Pty pty ->
		 ( match pty.ptystate with
		     | `Connected id -> if id <> handle.ct_id then raise Not_found
		     | _ -> raise Not_found
		 );
		 set_connect_timer ues name handle.ct_id at pty;
		 if do_output pty () then (
		   pty.ptyoutch # request_notification (do_output pty);
		   Unixqueue.once ues g (Int32.to_float timeout +. 0.1) pty.ptyinch#notify;
		 )
	 )
       with
	 | Not_found ->
	     reply (`code_error "Terminal not found or not connected")
    )
    session () reply
;;

let proc_disconnect ues gs =
  protect
    (fun session (handle,forced) reply ->
       Umlmon_auth.ensure_authenticated session;
       let umlvm = gs.umlvm in
       try
	 let at = find_term umlvm handle.ct_name in  (* or Not_found *)
	 ( match at.actv_termstatus with
	     | `Pty pty ->
 		 ( match pty.ptystate with
		     | `Connected id 
		     | `Connected_with(id,_,_) ->
			 if not forced && id <> handle.ct_id then raise Not_found
		     | _ -> raise Not_found
		 );
		 do_disconnect ues at pty
	 );
	 reply `code_ok
       with
	 | Not_found ->
	     reply (`code_error "Terminal not found or not connected")
    )
;;

let proc_is_connected ues gs =  
  protect
    (fun session name reply ->
       Umlmon_auth.ensure_authenticated session;
       let umlvm = gs.umlvm in
       try
	 let at = find_term umlvm name in  (* or Not_found *)
	 ( match at.actv_termstatus with
	     | `Pty pty ->
 		 ( match pty.ptystate with
		     | `Connected _ -> ()
		     | `Connected_with _ -> ()
		     | _ -> raise Not_found
		 );
		 reply true
	 )
       with 
	   Not_found -> reply false
    )
;;


let proc_te_init ues gs =
  protect
    (fun session (handle,width,height) reply ->
       Umlmon_auth.ensure_authenticated session;
       let umlvm = gs.umlvm in
       try
	 let at = find_term umlvm handle.ct_name in  (* or Not_found *)
	 ( match at.actv_termstatus with
	     | `Pty pty ->
 		 ( match pty.ptystate with
		     | `Connected id ->
			 if handle.ct_id <> id then raise Not_found
		     | _ -> 
			 raise Not_found
		 );
	 );
	 if width < 1l || height < 1l || (height > (Int32.div 20000l width)) then
	   failwith "Dimensions too large";
	 let w = Int32.to_int width in
	 let h = Int32.to_int height in
	 at.actv_termemu <- Some (new Termemu.terminal_emulator w h);
	 reply `code_ok
       with 
	 | Not_found -> 
	     reply (`code_error "Terminal not found or not connected")
	 | Failure msg ->
	     reply (`code_error msg)
    )
;;


let proc_te_read ues gs =
  protect 
    (fun session (handle,timeout) reply ->
       let umlvm = gs.umlvm in
       try
	 let at = find_term umlvm handle.ct_name in  (* or Not_found *)
	 match at.actv_termemu with
	   | None ->
	       failwith "No terminal emulation initialized - call te_init first";
	   | Some te ->
	       proc_read ues gs session (handle,timeout)
		 (function
		    | `code_ok data ->
			for k = 0 to String.length data - 1 do
			  te # output_char data.[k]
			done;
			reply `code_ok
		    | `code_interrupt ->
			reply `code_interrupt
		    | `code_in_progress ->
			assert false
		    | `code_error msg ->
			reply (`code_error msg)
		 )
       with 
	 | Not_found -> 
	     reply (`code_error "Terminal not found or not connected")
	 | Failure msg ->
	     reply (`code_error msg)
    )

;;


let tr_named_key =
  function
    | "UP"         -> `Cursor_up
    | "DOWN"       -> `Cursor_down
    | "LEFT"       -> `Cursor_left
    | "RIGHT"      -> `Cursor_right
    | "PGUP"       -> `Page_up
    | "PGDOWN"     -> `Page_down
    | "HOME"       -> `Home
    | "END"        -> `End
    | "INSERT"     -> `Insert
    | "DELETE"     -> `Delete
    | "BS"         -> `Backspace
    | "BACKSPACE"  -> `Backspace
    | "ENTER"      -> `Enter
    | "TAB"        -> `Tab
    | "ESC"        -> `Escape
    | "ESCAPE"     -> `Escape
    | "F1"         -> `Fn 1
    | "F2"         -> `Fn 2
    | "F3"         -> `Fn 3
    | "F4"         -> `Fn 4
    | "F5"         -> `Fn 5
    | "F6"         -> `Fn 6
    | "F7"         -> `Fn 7
    | "F8"         -> `Fn 8
    | "F9"         -> `Fn 9
    | "F10"        -> `Fn 10
    | "F11"        -> `Fn 11
    | "F12"        -> `Fn 12
    | _            -> failwith "Key name is not known"
;;


let proc_te_write_keys ues gs =
  protect 
    (fun session (handle,keys,timeout) reply ->
       let umlvm = gs.umlvm in
       try
	 let at = find_term umlvm handle.ct_name in  (* or Not_found *)
	 match at.actv_termemu with
	   | None ->
	       failwith "No terminal emulation initialized - call te_init first";
	   | Some te ->
	       Array.iter
		 (fun key ->
		    let k = 
		      match key with
			| `key_translated code -> `Translated (Int32.to_int code)
			| `key_cursor_up       -> `Cursor_up
			| `key_cursor_down     -> `Cursor_down
			| `key_cursor_left     -> `Cursor_left 
			| `key_cursor_right    -> `Cursor_right
			| `key_page_up         -> `Page_up  
			| `key_page_down       -> `Page_down
			| `key_home            -> `Home
			| `key_end             -> `End 
			| `key_insert          -> `Insert
			| `key_delete          -> `Delete
			| `key_backspace       -> `Backspace
			| `key_enter           -> `Enter
			| `key_tab             -> `Tab
			| `key_escape          -> `Escape
			| `key_fn nr           -> `Fn (Int32.to_int nr)
			| `key_named name      -> tr_named_key name
		    in
		    te # add_input_key k
		 )
		 keys;
	       let data = te # input_string() in
	       proc_write ues gs session (handle,data,timeout) 
		 (function
		    | `code_ok          -> reply `code_ok
		    | `code_error msg   -> reply (`code_error msg)
		    | `code_interrupt _ -> reply `code_interrupt
		    | `code_in_progress -> assert false
		 )
       with 
	 | Not_found -> 
	     reply (`code_error "Terminal not found or not connected")
	 | Failure msg ->
	     reply (`code_error msg)
    )

;;


let pack_cell (char, atts) =
  let i4 =  Rtypes.int_of_int4 in
  let u4 =  Rtypes.int_of_uint4 in
  let color_bits =
    function
      | Black   -> i4 c_black
      | Red     -> i4 c_red
      | Green   -> i4 c_green
      | Brown   -> i4 c_brown
      | Blue    -> i4 c_blue
      | Magenta -> i4 c_magenta
      | Cyan    -> i4 c_cyan
      | White   -> i4 c_white
  in
  let cell_char =
    match char with
      | Empty -> (-1)
      | Unicode n -> n in
  let cell_atts =
    ( match atts.intense with
	| I_dim  when char <> Empty -> (i4 i_dim) lsl (u4 shift_intensity)
	| I_bold when char <> Empty -> (i4 i_bold) lsl (u4 shift_intensity)
	| _                         -> (i4 i_norm) lsl (u4 shift_intensity)
    ) lor
    ( match atts.underscore with
	| false -> (i4 u_off) lsl (u4 shift_underscore)
	| true  -> (i4 u_on) lsl (u4 shift_underscore)
    ) lor
    ( match atts.blink with
	| true  when char <> Empty -> (i4 b_on) lsl (u4 shift_blink)
	| _                        -> (i4 b_off) lsl (u4 shift_blink)
    ) lor
    ( match atts.reverse with
	| true  when char <> Empty -> (i4 r_on) lsl (u4 shift_reverse)
	| _                        -> (i4 r_off) lsl (u4 shift_reverse)
    ) lor
    ( (color_bits atts.fg_color) lsl (u4 shift_fg_color) ) lor
    ( (color_bits atts.bg_color) lsl (u4 shift_bg_color) ) in

  { cell_char = Int32.of_int cell_char;
    cell_atts = Int32.of_int cell_atts
  }
;;


let proc_te_state ues gs =
  protect 
    (fun session (handle) reply ->
       let umlvm = gs.umlvm in
       try
	 let at = find_term umlvm handle.ct_name in  (* or Not_found *)
	 match at.actv_termemu with
	   | None ->
	       failwith "No terminal emulation initialized - call te_init first";
	   | Some te ->
	       let r = 
		 { te_height = Int32.of_int te#height;
		   te_width = Int32.of_int te#width;
		   te_cursor_x = Int32.of_int te#cursor_x;
		   te_cursor_y = Int32.of_int te#cursor_y;
		   te_cursor_visible = te#cursor_visible;
		   te_screen = Array.map (fun r -> Array.map pack_cell r) te#screen
		 } in
	       reply (Some r)
       with 
	 | Not_found -> 
	     reply None
	 | Failure msg ->
	     reply None
    )
;;


let proc_te_reset ues gs =
  protect 
    (fun session (handle) reply ->
       let umlvm = gs.umlvm in
       try
	 let at = find_term umlvm handle.ct_name in  (* or Not_found *)
	 match at.actv_termemu with
	   | None ->
	       failwith "No terminal emulation initialized - call te_init first";
	   | Some te ->
	       te # reset();
	       reply `code_ok
       with 
	 | Not_found -> 
	     reply (`code_error "Terminal not found or not connected")
	 | Failure msg ->
	     reply (`code_error msg)
    )
;;

